
//
//	Show how to display an irregular shaped button with transparent areas leaving the
//	background untouched.
//
//	Mark C. Williston ( http://www.williston.ca )
//  You have my permission to use this code as you see fit.
//
//	Uses Jean-Edouard Lachand-Robert's BitmapToRegion function
//  Give Jean-Edouard credit if you use this source and his function
//

#include <windows.h>
#include "resource.h"

#define GREENBUTTON 45001
#define BLUEBUTTON 45002
#define XBUTTON 45003

char szClassName[] = "Odd Buttonz"; // Irregular Shaped Button Region Example
char szNewButtonClassName[] = "newbutton";

HINSTANCE hInst;
HWND hWndMain;

HWND hWndGreenBtn;
HWND hWndBlueBtn;
HWND hWndXCloseBtn;

HBITMAP hBmpSkin;
HBITMAP hBmpGreenBtnUp;
HBITMAP hBmpGreenBtnDn;
HBITMAP hBmpGreenBtnMask;
HBITMAP hBmpBlueBtnUp;
HBITMAP hBmpBlueBtnDn;
HBITMAP hBmpBlueBtnMask;
HBITMAP hBmpXCloseBtnUp;
HBITMAP hBmpXCloseBtnDn;
HBITMAP hBmpXCloseBtnMask;

HRGN hRgnMain;
HRGN hRgnGreenButton;
HRGN hRgnBlueButton;
HRGN hRgnXCloseButton;

int iTimeToExit = 0;

//
//	BitmapToRegion : Create a region from the "non-transparent" pixels of a bitmap
//	Author : Jean-Edouard Lachand-Robert (http://www.geocities.com/Paris/LeftBank/1160/resume.htm), June 1998.
//
//	hBmp : Source bitmap
//	cTransparentColor : Color base for the "transparent" pixels (default is black)
//	cTolerance : Color tolerance for the "transparent" pixels.
//
//	A pixel is assumed to be transparent if the value of each of its 3 components (blue, green and red) is 
//	greater or equal to the corresponding value in cTransparentColor and is lower or equal to the 
//	corresponding value in cTransparentColor + cTolerance.
//

HRGN BitmapToRegion (HBITMAP hBmp, 	COLORREF cTransparentColor = 0, COLORREF cTolerance = 0x000000)
{
	HRGN hRgn = NULL;

	if (hBmp)
	{
		// Create a memory DC inside which we will scan the bitmap content
		HDC hMemDC = CreateCompatibleDC(NULL);
		if (hMemDC)
		{
			// Get bitmap size
			BITMAP bm;
			GetObject(hBmp, sizeof(bm), &bm);

			// Create a 32 bits depth bitmap and select it into the memory DC 
			BITMAPINFOHEADER RGB32BITSBITMAPINFO = {	
					sizeof(BITMAPINFOHEADER),	// biSize 
					bm.bmWidth,					// biWidth; 
					bm.bmHeight,				// biHeight; 
					1,							// biPlanes; 
					32,							// biBitCount 
					BI_RGB,						// biCompression; 
					0,							// biSizeImage; 
					0,							// biXPelsPerMeter; 
					0,							// biYPelsPerMeter; 
					0,							// biClrUsed; 
					0							// biClrImportant; 
			};
			VOID * pbits32; 
			HBITMAP hbm32 = CreateDIBSection(hMemDC, (BITMAPINFO *)&RGB32BITSBITMAPINFO, DIB_RGB_COLORS, &pbits32, NULL, 0);
			if (hbm32)
			{
				HBITMAP holdBmp = (HBITMAP)SelectObject(hMemDC, hbm32);

				// Create a DC just to copy the bitmap into the memory DC
				HDC hDC = CreateCompatibleDC(hMemDC);
				if (hDC)
				{
					// Get how many bytes per row we have for the bitmap bits (rounded up to 32 bits)
					BITMAP bm32;
					GetObject(hbm32, sizeof(bm32), &bm32);
					while (bm32.bmWidthBytes % 4)
						bm32.bmWidthBytes++;

					// Copy the bitmap into the memory DC
					HBITMAP holdBmp = (HBITMAP)SelectObject(hDC, hBmp);
					BitBlt(hMemDC, 0, 0, bm.bmWidth, bm.bmHeight, hDC, 0, 0, SRCCOPY);

					// For better performances, we will use the ExtCreateRegion() function to create the
					// region. This function take a RGNDATA structure on entry. We will add rectangles by
					// amount of ALLOC_UNIT number in this structure.
					#define ALLOC_UNIT	100
					DWORD maxRects = ALLOC_UNIT;
					HANDLE hData = GlobalAlloc(GMEM_MOVEABLE, sizeof(RGNDATAHEADER) + (sizeof(RECT) * maxRects));
					RGNDATA *pData = (RGNDATA *)GlobalLock(hData);
					pData->rdh.dwSize = sizeof(RGNDATAHEADER);
					pData->rdh.iType = RDH_RECTANGLES;
					pData->rdh.nCount = pData->rdh.nRgnSize = 0;
					SetRect(&pData->rdh.rcBound, MAXLONG, MAXLONG, 0, 0);

					// Keep on hand highest and lowest values for the "transparent" pixels
					BYTE lr = GetRValue(cTransparentColor);
					BYTE lg = GetGValue(cTransparentColor);
					BYTE lb = GetBValue(cTransparentColor);
					BYTE hr = min(0xff, lr + GetRValue(cTolerance));
					BYTE hg = min(0xff, lg + GetGValue(cTolerance));
					BYTE hb = min(0xff, lb + GetBValue(cTolerance));

					// Scan each bitmap row from bottom to top (the bitmap is inverted vertically)
					BYTE *p32 = (BYTE *)bm32.bmBits + (bm32.bmHeight - 1) * bm32.bmWidthBytes;
					for (int y = 0; y < bm.bmHeight; y++)
					{
						// Scan each bitmap pixel from left to right
						for (int x = 0; x < bm.bmWidth; x++)
						{
							// Search for a continuous range of "non transparent pixels"
							int x0 = x;
							LONG *p = (LONG *)p32 + x;
							while (x < bm.bmWidth)
							{
								BYTE b = GetRValue(*p);
								if (b >= lr && b <= hr)
								{
									b = GetGValue(*p);
									if (b >= lg && b <= hg)
									{
										b = GetBValue(*p);
										if (b >= lb && b <= hb)
											// This pixel is "transparent"
											break;
									}
								}
								p++;
								x++;
							}
							if (x > x0)
							{
								// Add the pixels (x0, y) to (x, y+1) as a new rectangle in the region
								if (pData->rdh.nCount >= maxRects)
								{
									GlobalUnlock(hData);
									maxRects += ALLOC_UNIT;
									hData = GlobalReAlloc(hData, sizeof(RGNDATAHEADER) + (sizeof(RECT) * maxRects), GMEM_MOVEABLE);
									pData = (RGNDATA *)GlobalLock(hData);
								}
								RECT *pr = (RECT *)&pData->Buffer;
								SetRect(&pr[pData->rdh.nCount], x0, y, x, y+1);
								if (x0 < pData->rdh.rcBound.left)
									pData->rdh.rcBound.left = x0;
								if (y < pData->rdh.rcBound.top)
									pData->rdh.rcBound.top = y;
								if (x > pData->rdh.rcBound.right)
									pData->rdh.rcBound.right = x;
								if (y+1 > pData->rdh.rcBound.bottom)
									pData->rdh.rcBound.bottom = y+1;
								pData->rdh.nCount++;
								// On Windows98, ExtCreateRegion() may fail if the number of rectangles is too
								// large (ie: > 4000). Therefore, we have to create the region by multiple steps.
								if (pData->rdh.nCount == 2000)
								{
									HRGN h = ExtCreateRegion(NULL, sizeof(RGNDATAHEADER) + (sizeof(RECT) * maxRects), pData);
									if (hRgn)
									{
										CombineRgn(hRgn, hRgn, h, RGN_OR);
										DeleteObject(h);
									}
									else
										hRgn = h;
									pData->rdh.nCount = 0;
									SetRect(&pData->rdh.rcBound, MAXLONG, MAXLONG, 0, 0);
								}
							}
						}
						// Go to next row (remember, the bitmap is inverted vertically)
						p32 -= bm32.bmWidthBytes;
					}
					// Create or extend the region with the remaining rectangles
					HRGN h = ExtCreateRegion(NULL, sizeof(RGNDATAHEADER) + (sizeof(RECT) * maxRects), pData);
					if (hRgn)
					{
						CombineRgn(hRgn, hRgn, h, RGN_OR);
						DeleteObject(h);
					}
					else
						hRgn = h;
					// Clean up
					GlobalFree(hData);
					SelectObject(hDC, holdBmp);
					DeleteDC(hDC);
				}
				DeleteObject(SelectObject(hMemDC, holdBmp));
			}
			DeleteDC(hMemDC);
		}	
	}
	return hRgn;
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	PAINTSTRUCT ps;
	HDC hdc;
	BITMAP bm;
	HDC memdc;
	HBITMAP h;
	HDC hdcMem; 
	LPDRAWITEMSTRUCT lpdis; 
	BITMAP bmPlay;
	HBITMAP hBmpBtnMask;
	WNDCLASS ClassNewButton;
	POINT ptTopLeft;
	RECT rcWindow;
	BOOL retValue;
	RECT rcDesktopWindow;
	long width;
	long height;

	switch (message) 
	{

		case WM_CREATE:

			// Center window across and move down one third
			GetWindowRect (hWnd, &rcWindow);
			width = rcWindow.right - rcWindow.left;
			height = rcWindow.bottom - rcWindow.top;
            
			GetWindowRect (GetDesktopWindow(), &rcDesktopWindow);
			ptTopLeft.x = (short)((rcDesktopWindow.right + rcDesktopWindow.left) / 2 - width / 2);
			ptTopLeft.y = (short)((rcDesktopWindow.top + rcDesktopWindow.bottom) / 3 - height / 3);
			retValue = SetWindowPos (hWnd, 0, ptTopLeft.x, ptTopLeft.y, 0, 0, SWP_NOZORDER | SWP_NOSIZE);

			// Make our new button class including the hand cursor
			if( GetClassInfo(	hInst, "button", &ClassNewButton ))
			{
				ClassNewButton.hCursor = LoadCursor(NULL, IDC_HAND);
				ClassNewButton.lpszClassName = szNewButtonClassName;
				ClassNewButton.style &= ~((UINT)CS_DBLCLKS);// !!! turn off CS_DBLCLKS style
				RegisterClass(&ClassNewButton);
			}

			// Green Button
			hBmpGreenBtnUp = (HBITMAP)LoadImage(hInst, MAKEINTRESOURCE(IDB_GREENBUTTON_UP), IMAGE_BITMAP, 0, 0, LR_CREATEDIBSECTION);
			hBmpGreenBtnDn = (HBITMAP)LoadImage(hInst, MAKEINTRESOURCE(IDB_GREENBUTTON_DN), IMAGE_BITMAP, 0, 0, LR_CREATEDIBSECTION);
			hBmpGreenBtnMask = (HBITMAP)LoadImage(hInst, MAKEINTRESOURCE(IDB_GREENBUTTON_MASK), IMAGE_BITMAP, 0, 0, LR_MONOCHROME);
			// Only need the up button to make the region
			if (hBmpGreenBtnUp)
			{
				GetObject(hBmpGreenBtnUp, sizeof(bmPlay), &bmPlay);
				hWndGreenBtn = CreateWindow(szNewButtonClassName, "Green Button", WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON | BS_OWNERDRAW, 21, 21, bmPlay.bmWidth, bmPlay.bmHeight, hWnd, (HMENU) GREENBUTTON, hInst, NULL);
				hRgnGreenButton = BitmapToRegion(hBmpGreenBtnUp);
				if (hRgnGreenButton)
					SetWindowRgn(hWndGreenBtn, hRgnGreenButton, TRUE);
				ShowWindow(hWndGreenBtn, SW_SHOW);
				UpdateWindow(hWndGreenBtn);				
			}

			// Same for the Blue Button
			hBmpBlueBtnUp = (HBITMAP)LoadImage(hInst, MAKEINTRESOURCE(IDB_BLUEBUTTON_UP), IMAGE_BITMAP, 0, 0, LR_CREATEDIBSECTION);
			hBmpBlueBtnDn = (HBITMAP)LoadImage(hInst, MAKEINTRESOURCE(IDB_BLUEBUTTON_DN), IMAGE_BITMAP, 0, 0, LR_CREATEDIBSECTION);
			hBmpBlueBtnMask = (HBITMAP)LoadImage(hInst, MAKEINTRESOURCE(IDB_BLUEBUTTON_MASK), IMAGE_BITMAP, 0, 0, LR_MONOCHROME);
			// And for the region
			if (hBmpBlueBtnUp)
			{
				GetObject(hBmpBlueBtnUp, sizeof(bmPlay), &bmPlay);
				hWndBlueBtn = CreateWindow(szNewButtonClassName, "Blue Button", WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON | BS_OWNERDRAW, 45, 45, bmPlay.bmWidth, bmPlay.bmHeight, hWnd, (HMENU) BLUEBUTTON, hInst, NULL);
				hRgnBlueButton = BitmapToRegion(hBmpBlueBtnUp);
				if (hRgnBlueButton)
					SetWindowRgn(hWndBlueBtn, hRgnBlueButton, TRUE);
				ShowWindow(hWndBlueBtn, SW_SHOW);
				UpdateWindow(hWndBlueBtn);
			}

			// Same for the X Button
			hBmpXCloseBtnUp = (HBITMAP)LoadImage(hInst, MAKEINTRESOURCE(IDB_XBUTTON_UP), IMAGE_BITMAP, 0, 0, LR_CREATEDIBSECTION);
			hBmpXCloseBtnDn = (HBITMAP)LoadImage(hInst, MAKEINTRESOURCE(IDB_XBUTTON_DN), IMAGE_BITMAP, 0, 0, LR_CREATEDIBSECTION);
			hBmpXCloseBtnMask = (HBITMAP)LoadImage(hInst, MAKEINTRESOURCE(IDB_XBUTTON_MASK), IMAGE_BITMAP, 0, 0, LR_MONOCHROME);
			// And for the region
			if (hBmpBlueBtnUp)
			{
				GetObject(hBmpXCloseBtnUp, sizeof(bmPlay), &bmPlay);
				hWndXCloseBtn = CreateWindow(szNewButtonClassName, "XClose Button", WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON | BS_OWNERDRAW, 98, 8, bmPlay.bmWidth, bmPlay.bmHeight, hWnd, (HMENU) XBUTTON, hInst, NULL);
				hRgnXCloseButton = BitmapToRegion(hBmpXCloseBtnUp);
				if (hRgnXCloseButton)
					SetWindowRgn(hWndXCloseBtn, hRgnXCloseButton, TRUE);
				ShowWindow(hWndXCloseBtn, SW_SHOW);
				UpdateWindow(hWndXCloseBtn);
			}
			return 0;

		// Allows us to move the region window
		case WM_LBUTTONDOWN:
			SendMessage(hWnd, WM_NCLBUTTONDOWN, HTCAPTION,NULL);
			return 0;

		// Needed for cases such as Hide Desktop/Show Desktop
		// Otherwise you may get a dirty button background
		case WM_CTLCOLORBTN:
			return (LRESULT)GetStockObject(HOLLOW_BRUSH);

		// Our code to draw the Owner Drawn Button
		case WM_DRAWITEM: 
			lpdis = (LPDRAWITEMSTRUCT) lParam; 
			hdcMem = CreateCompatibleDC(lpdis->hDC); 

			if (lpdis->itemState & ODS_SELECTED)  // if selected 
			{
				// A button was pressed
				switch(lpdis->CtlID)
				{
					case GREENBUTTON:
						SelectObject(hdcMem, hBmpGreenBtnDn); 
						hBmpBtnMask = hBmpGreenBtnMask;
						break;

					case BLUEBUTTON:
						SelectObject(hdcMem, hBmpBlueBtnDn); 
						hBmpBtnMask = hBmpBlueBtnMask;
						break;

					case XBUTTON:
						SelectObject(hdcMem, hBmpXCloseBtnDn); 
						hBmpBtnMask = hBmpXCloseBtnMask;
						iTimeToExit = 1;
						break;

					default:
						DeleteDC(hdcMem); 
						return TRUE;
				}
			}
			else 
			{
				//Button was released, lost focus or just needs to be drawn
				switch(lpdis->CtlID)
				{
					case GREENBUTTON:
						SelectObject(hdcMem, hBmpGreenBtnUp); 
						hBmpBtnMask = hBmpGreenBtnMask;
						break;

					case BLUEBUTTON:
						SelectObject(hdcMem, hBmpBlueBtnUp); 
						hBmpBtnMask = hBmpBlueBtnMask;
						break;

					case XBUTTON:
						SelectObject(hdcMem, hBmpXCloseBtnUp); 
						hBmpBtnMask = hBmpXCloseBtnMask;
						break;

					default:
						DeleteDC(hdcMem); 
						return TRUE;
				}
			}

			// BitBlt will not work correctly with irregular shaped buttons
			// as it erases the RECT region of the button
			// But MaskBlt works fine with a correct ROP and Mask Bitmap.

			MaskBlt(lpdis->hDC, 
				lpdis->rcItem.left, // x upper left 
				lpdis->rcItem.top,  // y upper left 
				lpdis->rcItem.right - lpdis->rcItem.left, 
				lpdis->rcItem.bottom - lpdis->rcItem.top, 
				hdcMem, 0, 0, hBmpBtnMask, 0, 0, 
				MAKEROP4(SRCPAINT,SRCCOPY));

			DeleteDC(hdcMem); 
			return TRUE; 

		case WM_PAINT:
				// Paint our Main Region Window
				hdc = BeginPaint (hWnd, &ps);
				if (hBmpSkin)
				{
					GetObject(hBmpSkin, sizeof(bm), &bm);
					memdc = CreateCompatibleDC(NULL);
					h = (HBITMAP)SelectObject(memdc, hBmpSkin);
					BitBlt(hdc, 0, 0, bm.bmWidth, bm.bmHeight, memdc, 0, 0, SRCCOPY);
					SelectObject(memdc, h);
				}
				EndPaint (hWnd, &ps);
			return 0;

		// Self explainatory
		case WM_COMMAND:
			switch(HIWORD(wParam))
			{
				case BN_CLICKED:
					switch(LOWORD(wParam))
					{
						case XBUTTON:
							SendMessage(hWndMain, WM_DESTROY, 0, 0);
							return 0;

						case GREENBUTTON:
							break;

						case BLUEBUTTON:
							break;
					}

				default:
					break;
			}
			break;

		case WM_DESTROY:
			// Delete what we loaded from memory
			DeleteObject(hBmpSkin);
			DeleteObject(hBmpGreenBtnUp);
			DeleteObject(hBmpGreenBtnDn);
			DeleteObject(hBmpGreenBtnMask);
			DeleteObject(hBmpBlueBtnUp);
			DeleteObject(hBmpBlueBtnDn);
			DeleteObject(hBmpBlueBtnMask);
			DeleteObject(hBmpXCloseBtnUp);
			DeleteObject(hBmpXCloseBtnDn);
			DeleteObject(hBmpXCloseBtnMask);
			DeleteObject(hRgnMain);
			DeleteObject(hRgnGreenButton);
			DeleteObject(hRgnBlueButton);
			DeleteObject(hRgnXCloseButton);

			// End it all here.
			PostQuitMessage(0);
			return 0;

		default:
			break;
	}
	return (DefWindowProc(hWnd, message, wParam, lParam));
}

int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow)
{

	WNDCLASS wc;
	MSG msg;
	BITMAP bm;

	hInst = hInstance;
	hBmpSkin = (HBITMAP)LoadImage(hInstance, MAKEINTRESOURCE(IDB_MAIN_SKIN), IMAGE_BITMAP, 0, 0, LR_CREATEDIBSECTION);
	if (hBmpSkin)
	{
		ZeroMemory(&wc, sizeof(wc));
		wc.style         = CS_HREDRAW | CS_VREDRAW;
		wc.lpszClassName = szClassName;
		wc.lpfnWndProc   = (WNDPROC)WndProc;
		wc.hInstance     = hInstance;
		wc.hCursor = LoadCursor(NULL,IDC_ARROW);	
		wc.hIcon = (HICON) LoadImage(hInstance, MAKEINTRESOURCE(IDI_ICON1), IMAGE_ICON, 32, 32, LR_DEFAULTSIZE);

		if (RegisterClass(&wc))
		{
			GetObject(hBmpSkin, sizeof(bm), &bm);

			// Create main window
			hWndMain = CreateWindow(szClassName, szClassName, WS_POPUP | WS_CLIPCHILDREN , CW_USEDEFAULT, CW_USEDEFAULT, bm.bmWidth, bm.bmHeight, NULL, NULL, hInstance, NULL);
			if (hWndMain)
			{
				// Set the skin bitmap to the region of the main window 
				hRgnMain = BitmapToRegion(hBmpSkin);
				if (hRgnMain)
					SetWindowRgn(hWndMain, hRgnMain, TRUE);
				ShowWindow(hWndMain, nCmdShow);
				UpdateWindow(hWndMain);

				// Main message pump:
				while (GetMessage(&msg, NULL, 0, 0)) 
				{
					TranslateMessage(&msg);
					DispatchMessage(&msg);
				}
				return (msg.wParam);
			}
		}
	}
	return -1;
}
